﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2011 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.PermutationEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.QuadraticAssignment {
  [StorableClass]
  public class QAPEvaluator : SingleSuccessorOperator, IQAPEvaluator {

    public ILookupParameter<Permutation> PermutationParameter {
      get { return (ILookupParameter<Permutation>)Parameters["Permutation"]; }
    }
    public ILookupParameter<BoolValue> UseDistanceMatrixParameter {
      get { return (ILookupParameter<BoolValue>)Parameters["UseDistanceMatrix"]; }
    }
    public ILookupParameter<DoubleMatrix> DistanceMatrixParameter {
      get { return (ILookupParameter<DoubleMatrix>)Parameters["DistanceMatrix"]; }
    }
    public ILookupParameter<DoubleMatrix> CoordinatesParameter {
      get { return (ILookupParameter<DoubleMatrix>)Parameters["Coordinates"]; }
    }
    public ILookupParameter<DoubleMatrix> WeightsParameter {
      get { return (ILookupParameter<DoubleMatrix>)Parameters["Weights"]; }
    }
    public ILookupParameter<DoubleValue> QualityParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["Quality"]; }
    }

    [StorableConstructor]
    protected QAPEvaluator(bool deserializing) : base(deserializing) { }
    protected QAPEvaluator(QAPEvaluator original, Cloner cloner) : base(original, cloner) { }
    public QAPEvaluator() {
      Parameters.Add(new LookupParameter<Permutation>("Permutation", "The permutation that represents the current solution."));
      Parameters.Add(new ValueLookupParameter<BoolValue>("UseDistanceMatrix", "True if the distance matrix should be used, false if the distances should be caluclated by taking the euclidean distance between the coordinates."));
      Parameters.Add(new LookupParameter<DoubleMatrix>("DistanceMatrix", "The distance matrix that contains the distances between the locations."));
      Parameters.Add(new LookupParameter<DoubleMatrix>("Coordinates", "The coordinates in case the distance matrix should not be used."));
      Parameters.Add(new LookupParameter<DoubleMatrix>("Weights", "The matrix with the weights between the facilities, that is how strongly they're connected to each other."));
      Parameters.Add(new LookupParameter<DoubleValue>("Quality", "The quality value aka fitness value of the solution."));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new QAPEvaluator(this, cloner);
    }

    public static double Apply(Permutation assignment, DoubleMatrix weights, Func<int, int, double> distance) {
      double quality = 0;
      for (int i = 0; i < assignment.Length; i++) {
        for (int j = 0; j < assignment.Length; j++) {
          quality += weights[i, j] * distance(assignment[i], assignment[j]);
        }
      }
      return quality;
    }

    public override IOperation Apply() {
      Permutation assignment = PermutationParameter.ActualValue;
      bool useDistanceMatrix = UseDistanceMatrixParameter.ActualValue.Value;
      DoubleMatrix weights = WeightsParameter.ActualValue;
      double quality;

      if (useDistanceMatrix) {
        DoubleMatrix distanceMatrix = DistanceMatrixParameter.ActualValue;
        quality = Apply(assignment, weights, (x, y) => distanceMatrix[x, y]);
      } else {
        DoubleMatrix coordinates = CoordinatesParameter.ActualValue;
        quality = Apply(assignment, weights, (x, y) => Distance(coordinates, x, y));
      }

      QualityParameter.ActualValue = new DoubleValue(quality);

      return base.Apply();
    }

    private double Distance(DoubleMatrix coordinates, int row1, int row2) {
      double dx = coordinates[row1, 0] - coordinates[row2, 0];
      double dy = coordinates[row1, 1] - coordinates[row2, 1];
      return Math.Sqrt(dx * dx + dy * dy);
    }
  }
}
